module QueryInterface
  module Server
    module Transformations

      def self.included(base)    
        base.extend(ClassMethods)
        base.dataset_module do
          def query_transformation(type, name, *args)
            block = model.instance_variable_get(:@query_transformations)[type][name]
            if block
              self.instance_exec(*args, &block)
            elsif type == :order and args.count == 1
              self.order_append(Sequel.send(args[0], name.to_sym))
            else
              raise "No #{type} implementation for #{name}"
            end
          end
        end
        base.instance_variable_set(:@query_transformations, {
          exclude: {},
          filter: {},
          with: {},
          order: {},
          update: {},
        })
      end

      module ClassMethods
        def query_transformations(&block)
          scope = TransformationScope.new
          scope.instance_eval(&block)

          @query_transformations[:exclude].merge!(scope.excludes)
          @query_transformations[:filter].merge!(scope.filters)
          @query_transformations[:with].merge!(scope.withs)
          @query_transformations[:order].merge!(scope.orders)
          @query_transformations[:update].merge!(scope.updates)
        end

        def query
          return QueryInterface::Server::LazyQuery.new(self)
        end

      end

      class TransformationScope
        attr_accessor :excludes, :filters, :withs, :orders, :updates

        def initialize
          self.excludes = {}
          self.filters = {}
          self.withs = {}
          self.orders = {}
          self.updates = {}
        end

        def auto_exclude(*names)
          names.each do |name|
            self.excludes[name] = Proc.new do |param|
              exclude(Sequel.qualify(model.table_name, name) => param)
            end
          end
        end

        def auto_filter(*names)
          names.each do |name|
            self.filters[name] = Proc.new do |param|
              filter(Sequel.qualify(model.table_name, name) => param)
            end
          end
        end

        def auto_update(*names)
          names.each do |name|
            self.updates[name] = Proc.new {|param| param}
          end
        end

        def exclude(name, &block)
          self.excludes[name] = block
        end

        def filter(name, &block)
          self.filters[name] = block
        end

        def update(name, &block)
          self.updates[name] = block
        end

        def with(name, &block)
          self.withs[name] = block
        end

        def order(name, &block)
          self.orders[name] = block
        end
      end

    end
  end
end
